import math

import torch
from torch import nn

from . import init
from .. import functional as C

Tensor = torch.Tensor


class ComplexLinear(nn.Module):
    def __init__(self, in_features: int, out_features: int, bias: bool = True):
        super().__init__()

        self.in_features = in_features
        self.out_features = out_features
        self.weight_real = nn.Parameter(torch.Tensor(out_features, in_features))
        self.weight_imag = nn.Parameter(torch.Tensor(out_features, in_features))
        if bias:
            self.bias = nn.Parameter(torch.Tensor(2 * out_features))
        else:
            self.bias = self.register_parameter("bias", None)
        self.reset_parameters()

    def reset_parameters(self):
        init.rayleigh_kaiming_(self.weight_real, self.weight_imag)
        if self.bias is not None:
            fan_in, _ = nn.init._calculate_fan_in_and_fan_out(self.weight_real)
            bound = 1 / math.sqrt(fan_in)
            nn.init.uniform_(self.bias, -bound, bound)

    def forward(self, x: Tensor) -> Tensor:
        return C.complex_linear(x, self.weight_real, self.weight_imag, self.bias)


__all__ = ["ComplexLinear"]



import math

import numpy as np
import torch
from torch import distributions

Tensor = torch.Tensor


def rayleigh_kaiming_(
    real: Tensor,
    imag: Tensor,
    a: float = 0,
    mode: str = "fan_in",
    nonlinearity: str = "leaky_relu",
) -> None:
    if real.shape != imag.shape:
        raise ValueError("real and imag tensors must have the same shape")

    fan = torch.nn.init._calculate_correct_fan(real, mode)
    gain = torch.nn.init.calculate_gain(nonlinearity, a)
    # Calculate Rayleigh scale parameter.
    # Derivation:
    # Var(W) = 2s^2
    # Var(W) = gain^2 / fan
    # s = sqrt(gain^2 / fan / 2) = gain / sqrt(fan) / sqrt(2)
    # See equation 10 in https://arxiv.org/pdf/1705.09792.pdf for more details.
    # NOTE: PyTorch does not have a built in Rayleigh distribution; however, the Weibull
    # distribution with shape parameter of 2 yields a Rayleigh distribution. The
    # Rayleigh scale parameter is related to the Weibull scale parameter via the
    # follwing equation: lambda = sigma * sqrt(2).
    # Given this, we can remove the final division by sqrt(2) from the derived equation
    # to calculate lambda directly.
    scale = gain / math.sqrt(fan)
    weibull = distributions.Weibull(scale, 2)  # concentration == shape == k
    uniform = distributions.Uniform(-np.pi, np.pi)
    with torch.no_grad():
        modulus = weibull.sample(real.shape)
        phase = uniform.sample(real.shape)
        # See equation 8 in the above paper.
        initialized_real = modulus * torch.cos(phase)
        initialized_imag = modulus * torch.sin(phase)
        real.copy_(initialized_real)
        imag.copy_(initialized_imag)


def rayleigh_xavier_(real: Tensor, imag: Tensor, gain: float = 1.0) -> None:
    if real.shape != imag.shape:
        raise ValueError("real and imag tensors must have the same shape")

    fan_in, fan_out = torch.nn.init._calculate_fan_in_and_fan_out(real)
    # Calculate Rayleigh scale parameter.
    # Derivation:
    # Var(W) = 2s^2
    # Var(W) = gain^2 * 2 / (fan_in + fan_out)
    # s = sqrt(gain^2 * 2 / (fan_in + fan_out) / 2) = gain * sqrt(1 / (fan_in + fan_out))
    # See equation 10 in https://arxiv.org/pdf/1705.09792.pdf for more details.
    # NOTE: PyTorch does not have a built in Rayleigh distribution; however, the Weibull
    # distribution with shape parameter of 2 yields a Rayleigh distribution. The
    # Rayleigh scale parameter is related to the Weibull scale parameter via the
    # follwing equation: lambda = sigma * sqrt(2).
    # Given this, we simply modify the derived equation by multiplying the contents of
    # the square root by 2.
    scale = gain * math.sqrt(2 / (fan_in + fan_out))
    weibull = distributions.Weibull(scale, 2)  # concentration == shape == k
    uniform = distributions.Uniform(-np.pi, np.pi)
    with torch.no_grad():
        modulus = weibull.sample(real.shape)
        phase = uniform.sample(real.shape)
        # See equation 8 in the above paper.
        initialized_real = modulus * torch.cos(phase)
        initialized_imag = modulus * torch.sin(phase)
        real.copy_(initialized_real)
        imag.copy_(initialized_imag)


__all__ = ["rayleigh_kaiming_", "rayleigh_xavier_"]



from typing import TypeVar, Union, Tuple

# NOTE: copied from latest master of PyTorch.
# Create some useful type aliases

# Template for arguments which can be supplied as a tuple, or which can be a scalar which PyTorch will internally
# broadcast to a tuple.
# Comes in several variants: A tuple of unknown size, and a fixed-size tuple for 1d, 2d, or 3d operations.
T = TypeVar('T')
_scalar_or_tuple_any_t = Union[T, Tuple[T, ...]]
_scalar_or_tuple_1_t = Union[T, Tuple[T]]
_scalar_or_tuple_2_t = Union[T, Tuple[T, T]]
_scalar_or_tuple_3_t = Union[T, Tuple[T, T, T]]

# For arguments which represent size parameters (eg, kernel size, padding)
_size_any_t = _scalar_or_tuple_any_t[int]
_size_1_t = _scalar_or_tuple_1_t[int]
_size_2_t = _scalar_or_tuple_2_t[int]
_size_3_t = _scalar_or_tuple_3_t[int]



import math

import torch
from torch import nn
from torch.nn import functional as F

from . import init
from .utils import _single, _pair, _triple, _reverse_repeat_tuple
from .. import common_types
from .. import functional as C

Tensor = torch.Tensor
_size_any_t = common_types._size_any_t
_size_1_t = common_types._size_1_t
_size_2_t = common_types._size_2_t
_size_3_t = common_types._size_3_t


class _ComplexConvNd(nn.Module):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: _size_any_t,
        stride: _size_any_t,
        padding: _size_any_t,
        dilation: _size_any_t,
        transposed: bool,
        output_padding: _size_any_t,
        groups: int,
        bias: bool,
        padding_mode: str,
    ):
        super().__init__()
        if in_channels % groups != 0:
            raise ValueError("in_channels must be divisible by groups")
        if out_channels % groups != 0:
            raise ValueError("out_channels must be divisible by groups")
        valid_padding_modes = {"zeros", "reflect", "replicate", "circular"}
        if padding_mode not in valid_padding_modes:
            raise ValueError(
                f"padding_mode must be one of {valid_padding_modes}, "
                f"but got padding_mode='{padding_mode}'"
            )
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.kernel_size = kernel_size
        self.stride = stride
        self.padding = padding
        self.dilation = dilation
        self.transposed = transposed
        self.output_padding = output_padding
        self.groups = groups
        self.padding_mode = padding_mode
        # `_reversed_padding_repeated_twice` is the padding to be passed to
        # `F.pad` if needed (e.g., for non-zero padding types that are
        # implemented as two ops: padding + conv). `F.pad` accepts paddings in
        # reverse order than the dimension.
        self._reversed_padding_repeated_twice = _reverse_repeat_tuple(self.padding, 2)
        if transposed:
            self.weight_real = nn.Parameter(
                torch.Tensor(in_channels, out_channels // groups, *kernel_size)
            )
            self.weight_imag = nn.Parameter(
                torch.Tensor(in_channels, out_channels // groups, *kernel_size)
            )
        else:
            self.weight_real = nn.Parameter(
                torch.Tensor(out_channels, in_channels // groups, *kernel_size)
            )
            self.weight_imag = nn.Parameter(
                torch.Tensor(out_channels, in_channels // groups, *kernel_size)
            )
        if bias:
            self.bias = nn.Parameter(torch.Tensor(2 * out_channels))
        else:
            self.register_parameter("bias", None)
        self.reset_parameters()

    def reset_parameters(self) -> None:
        init.rayleigh_kaiming_(self.weight_real, self.weight_imag)
        if self.bias is not None:
            fan_in, _ = nn.init._calculate_fan_in_and_fan_out(self.weight_real)
            bound = 1 / math.sqrt(fan_in)
            nn.init.uniform_(self.bias, -bound, bound)

    def extra_repr(self):
        s = (
            "{in_channels}, {out_channels}, kernel_size={kernel_size}"
            ", stride={stride}"
        )
        if self.padding != (0,) * len(self.padding):
            s += ", padding={padding}"
        if self.dilation != (1,) * len(self.dilation):
            s += ", dilation={dilation}"
        if self.output_padding != (0,) * len(self.output_padding):
            s += ", output_padding={output_padding}"
        if self.groups != 1:
            s += ", groups={groups}"
        if self.bias is None:
            s += ", bias=False"
        if self.padding_mode != "zeros":
            s += ", padding_mode={padding_mode}"
        return s.format(**self.__dict__)

    def __setstate__(self, state):
        super().__setstate__(state)
        if not hasattr(self, "padding_mode"):
            self.padding_mode = "zeros"


class ComplexConv1d(_ComplexConvNd):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: _size_1_t,
        stride: _size_1_t = 1,
        padding: _size_1_t = 0,
        dilation: _size_1_t = 1,
        groups: int = 1,
        bias: bool = True,
        padding_mode: str = "zeros",
    ):
        kernel_size = _single(kernel_size)
        stride = _single(stride)
        padding = _single(padding)
        dilation = _single(dilation)
        super().__init__(
            in_channels,
            out_channels,
            kernel_size,
            stride,
            padding,
            dilation,
            False,
            _single(0),
            groups,
            bias,
            padding_mode,
        )

    def forward(self, input: Tensor) -> Tensor:
        if self.padding_mode != "zeros":
            return C.complex_conv1d(
                F.pad(
                    input, self._reversed_padding_repeated_twice, mode=self.padding_mode
                ),
                self.weight_real,
                self.weight_imag,
                self.bias,
                self.stride,
                _single(0),
                self.dilation,
                self.groups,
            )
        return C.complex_conv1d(
            input,
            self.weight_real,
            self.weight_imag,
            self.bias,
            self.stride,
            self.padding,
            self.dilation,
            self.groups,
        )


class ComplexConv2d(_ComplexConvNd):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: _size_2_t,
        stride: _size_2_t = 1,
        padding: _size_2_t = 0,
        dilation: _size_2_t = 1,
        groups: int = 1,
        bias: bool = True,
        padding_mode: str = "zeros",
    ):
        kernel_size = _pair(kernel_size)
        stride = _pair(stride)
        padding = _pair(padding)
        dilation = _pair(dilation)
        super().__init__(
            in_channels,
            out_channels,
            kernel_size,
            stride,
            padding,
            dilation,
            False,
            _pair(0),
            groups,
            bias,
            padding_mode,
        )

    def forward(self, input: Tensor):
        if self.padding_mode != "zeros":
            return C.complex_conv2d(
                F.pad(
                    input, self._reversed_padding_repeated_twice, mode=self.padding_mode
                ),
                self.weight_real,
                self.weight_imag,
                self.bias,
                self.stride,
                _pair(0),
                self.dilation,
                self.groups,
            )
        return C.complex_conv2d(
            input,
            self.weight_real,
            self.weight_imag,
            self.bias,
            self.stride,
            self.padding,
            self.dilation,
            self.groups,
        )


class ComplexConv3d(_ComplexConvNd):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: _size_3_t,
        stride: _size_3_t = 1,
        padding: _size_3_t = 0,
        dilation: _size_3_t = 1,
        groups: int = 1,
        bias: bool = True,
        padding_mode: str = "zeros",
    ):
        kernel_size = _triple(kernel_size)
        stride = _triple(stride)
        padding = _triple(padding)
        dilation = _triple(dilation)
        super().__init__(
            in_channels,
            out_channels,
            kernel_size,
            stride,
            padding,
            dilation,
            False,
            _triple(0),
            groups,
            bias,
            padding_mode,
        )

    def forward(self, input: Tensor) -> Tensor:
        if self.padding_mode != "zeros":
            return C.complex_conv3d(
                F.pad(
                    input, self._reversed_padding_repeated_twice, mode=self.padding_mode
                ),
                self.weight_real,
                self.weight_imag,
                self.bias,
                self.stride,
                _triple(0),
                self.dilation,
                self.groups,
            )
        return C.complex_conv3d(
            input,
            self.weight_real,
            self.weight_imag,
            self.bias,
            self.stride,
            self.padding,
            self.dilation,
            self.groups,
        )


__all__ = ["ComplexConv1d", "ComplexConv2d", "ComplexConv3d"]



import collections
from itertools import repeat

# NOTE: copied from the PyTorch repository.


def _ntuple(n):
    def parse(x):
        if isinstance(x, collections.Iterable):
            return x
        return tuple(repeat(x, n))

    return parse


_single = _ntuple(1)
_pair = _ntuple(2)
_triple = _ntuple(3)
_quadruple = _ntuple(4)


def _reverse_repeat_tuple(t, n):
    r"""Reverse the order of `t` and repeat each element for `n` times.
    This can be used to translate padding arg used by Conv and Pooling modules
    to the ones used by `F.pad`.
    """
    return tuple(x for x in reversed(t) for _ in range(n))



from typing import Optional

import torch
from torch.nn import functional as F

from . import common_types

Tensor = torch.Tensor
_size_1_t = common_types._size_1_t
_size_2_t = common_types._size_2_t
_size_3_t = common_types._size_3_t


def complex_linear(
    input: Tensor, real: Tensor, imag: Tensor, bias: Optional[Tensor] = None
) -> Tensor:
    combined_real = torch.cat([real, -imag], dim=1)
    combined_imag = torch.cat([imag, real], dim=1)
    weight_complex = torch.cat([combined_real, combined_imag], dim=0)
    return F.linear(input, weight_complex, bias)


def complex_conv1d(
    input: Tensor,
    real: Tensor,
    imag: Tensor,
    bias: Optional[Tensor] = None,
    stride: _size_1_t = 1,
    padding: _size_1_t = 0,
    dilation: _size_1_t = 1,
    groups: int = 1,
) -> Tensor:
    combined_real = torch.cat([real, -imag], dim=1)
    combined_imag = torch.cat([imag, real], dim=1)
    weight_complex = torch.cat([combined_real, combined_imag], dim=0)
    return F.conv1d(input, weight_complex, bias, stride, padding, dilation, groups)


def complex_conv2d(
    input: Tensor,
    real: Tensor,
    imag: Tensor,
    bias: Optional[Tensor] = None,
    stride: _size_2_t = 1,
    padding: _size_2_t = 0,
    dilation: _size_2_t = 1,
    groups: int = 1,
) -> Tensor:
    combined_real = torch.cat([real, -imag], dim=1)
    combined_imag = torch.cat([imag, real], dim=1)
    weight_complex = torch.cat([combined_real, combined_imag], dim=0)
    return F.conv2d(input, weight_complex, bias, stride, padding, dilation, groups)


def complex_conv3d(
    input: Tensor,
    real: Tensor,
    imag: Tensor,
    bias: Tensor = None,
    stride: _size_3_t = 1,
    padding: _size_3_t = 0,
    dilation: _size_3_t = 1,
    groups: int = 1,
) -> Tensor:
    combined_real = torch.cat([real, -imag], dim=1)
    combined_imag = torch.cat([imag, real], dim=1)
    weight_complex = torch.cat([combined_real, combined_imag], dim=0)
    return F.conv3d(input, weight_complex, bias, stride, padding, dilation, groups)


__all__ = ["complex_conv1d", "complex_conv2d", "complex_conv3d", "complex_linear"]
